#!/usr/bin/env python
# 
# $Id: main.py,v 1.9 2001/06/18 09:43:53 drozd Exp $
# (c) Sergey Lyubka <devnull@tecnomen.ie>
#

import sys
import os.path
import os

# go home
if __name__ == '__main__':
    os.chdir(os.path.dirname(sys.argv[0]))
    sys.path.append(os.getcwd())



import struct
import socket
import select
import string
import time

import config
import const
import proto
import trans
import log
import xmlp
import buddy

from ui_tk import CUI


############################################################################
# classes
############################################################################
class CICQSession:

    def __init__(self, server, port):
        "the constructor"

        self.server    = server
        self.port      = port
        self.logger    = log.CLogger()
        self.root_dir  = config.get_root_dir()
        self.uin       = config.get_my_uin()
        trans_method   = config.get_translation_method()
        # now trans[0] decodes
        # trans[1] - encodes message
        self.trans     = trans.trans_table[trans_method]

        # contact list
        # it's keyed by __repr__() method of CBuddy() class
        # at the moment, this is buddy's nickname
        self.contacts    = {}
        self.contacts[self.uin] = buddy.CBuddy(self.uin, 'py-icq user')


        # init parser for incoming nessages
        self.parser    = xmlp.CParser()
        # user interface init
        self.ui        = CUI(self)
        self.pwd       = self.ui.get_passwd()
        for x in config.get_contacts().items():
            b = buddy.CBuddy(x[0], x[1])
            self.ui.contact_add(b)

        self.quit             = 0 # flag to quit from main loop

        self.logged_in        = 0 # are we logged in
        self.poll_sent        = 0
        self.start            = -1



    def main_loop(self, timeout=const.SELECT_TIMEOUT):
        # endless message dispatch loop
        while self.quit == 0:

            # poll if necessary
            if self.logged_in == 1 and \
            int((time.time()-self.start)) % const.POLL_DELAY == 0:
                if self.poll_sent == 0:
                    self.poll_server()
                    self.poll_sent = 1
            else:
                self.poll_sent = 0


            # get data from server
            i, o, e = select.select([self.sock], [], [], timeout )
            if len(i) > 0 :
                try:
                    data = self.sock.recv(const.TCP_PACKET_MAXLEN)
                except:
                    self.logged_in = 0
                    self.login()
                    return
                while len(data) != 0:
                    flap = proto.CFLAP(data)
                    if flap.channel == proto.AIM_FLAP_DATA:
                        snac = proto.CSNAC(flap.data)
                        self.handle_server_cmd(snac)
                    data = data[flap.data_len + 
                        flap.proto_headerlen:]


            # get data from user
            # we get a list of commands
            cmdlist = self.ui.get_input()
            if cmdlist is None:
                continue
            for entry in cmdlist:
                cmd, params = entry
                self.handle_user_cmd(cmd, params)

        self.disconnect()

    def poll_server(self):
        self.sock.send(proto.CFLAP(None, 
            proto.AIM_FLAP_CMD_START, 
            proto.AIM_FLAP_POLL, 0, '').packet )


    def handle_user_cmd(self, cmd, params):
        "process user input"
        if cmd not in const.cmd_table.keys():
            print 'unknown command %s' % cmd
            return

        # COMMAND SWITCH
        if cmd == 'clear':
            self.ui.clear()
        elif cmd == 'msg-out':
            uin, text = params
            msg = proto.CMsgOut(uin, text)
            # send the message
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_MSG,
                proto.AIM_SNAC_MSG_C_MSG, 0, 0, msg.packet)
            self.send_snac(p)



    def handle_server_cmd(self, snac):

        cmd = lookup_handler(snac)
        #print cmd

        if cmd == 'msg-in':
            m = proto.CMsgIn(snac.data)
            t = self.parser.strip_html(m.msg)
            if self.trans is not None:
                t = self.trans[0](t)
            self.logger.log_uin(m.uin, t)
            self.ui.msg_in(m.uin, self.uin, t)
        elif cmd == 'server-ready':
            self.start    = time.time()
            # send cryptic crap
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                23, 0, 0, proto.AIM_RAW_23)
            self.send_snac(p)

            # request rate information
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_REQ_RATE, 0, 0,'')
            self.send_snac(p)
        elif cmd == 'server-rate-response':

            # ack rate response 
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_RATE_ACK, 0, 0, 
                proto.AIM_RAW_RATE_ACK)
            self.send_snac(p)

            # request clientinfo
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_REQ_SCREEN_INFO, 0, 0, '')
            self.send_snac(p)

            # request rights info
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_SERV_LOC,
                proto.AIM_SNAC_LOC_C_REQ_PRIV, 0, 0, '')
            self.send_snac(p)

            # set user information
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_SERV_LOC,
                proto.AIM_SNAC_LOC_C_SET_USER_INFO, 0, 0, 
                tlv_make(1, 'text/x-aolrtf; charset="us-asci"') + \
                tlv_make(2, 'hello world')+ \
                tlv_make(3, 'text/x-aolrtf; charset="us-asci"' + \
                tlv_make(4, 'my away message')) + \
                tlv_make(5, '\x09\x46\x13\x46\x4c\x7f\x11\xd1' + \
                '\x82\x22\x44\x45\x53\x54\x00\x00' + \
                '\x09\x46\x13\x41\x4c\x7f\x11\xd1' + \
                '\x82\x22\x44\x45\x53\x54\x00\x00' + \
                '\x09\x46\x13\x45\x4c\x7f\x11\xd1' + \
                '\x82\x22\x44\x45\x53\x54\x00\x00' + \
                '\x74\x8f\x24\x20\x62\x87\x11\xd1' + \
                '\x82\x22\x44\x45\x53\x54\x00\x00') )
            self.send_snac(p)

            # request buddy list rights info
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_LIST,
                proto.AIM_SNAC_LST_C_REQ_PRIV, 0,0, '')
            self.send_snac(p)

            # add to buddy list
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_LIST,
                proto.AIM_SNAC_LST_C_ADD, 0,1, '\x04kuku')
            self.send_snac(p)

            # set user info
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_SERV_LOC,
                proto.AIM_SNAC_LOC_C_SET_USER_INFO, 0,0, 
                proto.AIM_RAW_SET_INFO)
            self.send_snac(p)

            # request messaging parameter info
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_MSG,
                proto.AIM_SNAC_MSG_C_REQ_PAR, 0,0, '') 
            self.send_snac(p)


            # request BOS rights
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_BOS,
                proto.AIM_SNAC_BOS_C_REQ_PRIV, 0,0, '') 
            self.send_snac(p)

            # set BOS permission mask
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_BOS,
                proto.AIM_SNAC_BOS_C_SET_GPERM, 0, 0,
                '\x1f')
            self.send_snac(p)

            # set privacy flags
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_SET_PRIVACY, 0, 0, 
                '\x00\x00\x00\x01')
            self.send_snac(p)

            #time.sleep(3)
            # mark ready
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_READY, 0, 0,
                proto.AIM_RAW_READY)
            self.send_snac(p)
        elif cmd == 'hash-request':
            # send hash response back
            p = proto.CSNAC(None,
                proto.AIM_SNAC_FAMILY_CTL,
                proto.AIM_SNAC_CTL_C_HASH_RESPONSE, 0, 0,
                proto.AIM_RAW_HASH)
            self.send_snac(p)

            # set logged in flag
            self.logged_in = 1

            # now, initialize our buddy list
            for b in self.contacts.values():
                if b.uin != self.uin:
                    self.add_to_buddylist(b)
        elif cmd == 'online':
            # mark contact as online
            print 'in online'
            b = proto.CBuddyOnline(snac.packet)
            self.ui.set_contact_mode(b.uin, const.ICQ_MODE_ONLINE)
        elif cmd == 'offline':
            print 'offline'
            b = proto.CBuddyOnline(snac.packet)
            self.ui.set_contact_mode(b.uin, const.ICQ_MODE_OFFLINE)
        else:
            print 'unexpected: family %d subtype %d' % (snac.family, snac.subtype)

        print cmd


    def send_snac(self, snac, flap_channel=proto.AIM_FLAP_DATA):
        "receive snac object, incapsulates into flap, and send to server"
        #self.ui.show(str(snac.packet))
        flap = proto.CFLAP(None, proto.AIM_FLAP_CMD_START,
            flap_channel, len(snac.packet), snac.packet)
        self.sock.send(flap.packet)


    def send_tlv_chain(self, chain, flap_channel=proto.AIM_FLAP_DATA):
        data = '\x00\x00\x00\x01'+chain
        self.sock.send(proto.CFLAP(None, proto.AIM_FLAP_CMD_START,
            flap_channel, len(data), data).packet)



    def login(self):
        self.connect(self.server,self.port)
        chain    = \
            tlv_make(1,self.uin) + \
            tlv_make(2,encode_password(self.pwd)) + \
            tlv_make(3,const.CLIENT_STRING) + \
            tlv_make(0x16, '\x01\x0a') + \
            tlv_make(0x17, '\x00\x04') + \
            tlv_make(0x18, '\x00\x3c') + \
            tlv_make(0x19, '\x00\x01') + \
            tlv_make(0x1a, '\x0c\x0e') + \
            tlv_make(0x14, '\x00\x00\x00\x55') + \
            tlv_make(0xe,'us') + tlv_make(0xf,'en')

        flap = proto.CFLAP( self.sock.recv(1024) )
        del flap


        # send auth request
        self.send_tlv_chain(chain, proto.AIM_FLAP_NEW)
        flap = proto.CFLAP(self.sock.recv(const.TCP_PACKET_MAXLEN))
        self.disconnect()

        tlv_chain = tlv_dissect_chain(flap.data)
        cookie = tlv_chain[proto.TLV_COOKIE]
        h,p = string.split(tlv_chain[proto.TLV_REDIR],':')

        self.connect(h,int(p))
        cookie_chain = tlv_make(proto.TLV_COOKIE, cookie)
        self.send_tlv_chain(cookie_chain, proto.AIM_FLAP_NEW )


    def connect(self,host,port):
        self.sock = socket.socket(socket.AF_INET, 
            socket.SOCK_STREAM, 0)
        self.sock.connect([host, port])


    def disconnect(self):
        self.sock.close()
        del self.sock


    def sleep(self):
        time.sleep(const.SELECT_TIMEOUT)


    def add_to_buddylist(self, b):
        '''
            send 'add to buddylist' command to server
            for the given buddy
            b is the CBuddy instance
        '''
        uin = b.uin
        snac = proto.CSNAC(None,
            proto.AIM_SNAC_FAMILY_LIST, proto.AIM_SNAC_LST_C_ADD,
            0, 0, hex(len(uin)) + uin )
        self.send_snac(snac)


############################################################################
# functions
############################################################################
def start(argv):
    server, port = config.get_server()

    conn = CICQSession(server=server, port=port)
    return
    conn.login()
    conn.main_loop()



def tlv_make(type, value):
    "return tlv-encoded string"
    l = len(value)
    fmt = '!hh %ds' % l
    return struct.pack(fmt, type, l, value)



def encode_password(pwd):
    '''
        encodes ICQ password
        look at XOR table used for encoding (const.py)
        every character of password XOR'ed
        with character from XOR table
    '''
    list = map(lambda x,y: chr(x^ord(y)), const.xor_table[:len(pwd)], pwd)
    return string.join(list,'')



def tlv_dissect_chain(data):
    '''
        dissect data, return dictionary hashed by TLV headers
    '''
    ret_val = {}
    while 1:
        if len(data) < 4:
            break
        id,l    = struct.unpack('!hh', data[:4])
        content    = data[4:(l+4)]
        ret_val[id] = content
        data    = data[4+l:]
    return ret_val



def lookup_handler(snac):
    "scans cmd_table for known commands"
    ret_val    = None
    for cmd in const.cmd_table.items():
        if cmd[1] == (snac.family,snac.subtype):
            ret_val = cmd[0]
            break
    return ret_val




###########################################################################
# the entry
############################################################################
if __name__ == '__main__':
    sys.exit(start(sys.argv))
