# $Id: proto.py,v 1.7 2001/06/18 09:43:53 drozd Exp $
# (c) 2001 Sergey M Lyubka <devnull@tecnomen.ie>
#

import struct
import string
import log




AIM_FLAP_CMD_START    = 0x2a

AIM_FLAP_NEW          = 0x01    # FLAP  channel ids
AIM_FLAP_DATA         = 0x02
AIM_FLAP_ERROR        = 0x03
AIM_FLAP_CLOSE        = 0x04
AIM_FLAP_POLL         = 0x05

#AIM_SNAC_SERVER_REQ_ID_MASK    = 2147483648    # server-initiated SNAC req_id is 
                        # ORed with this

AIM_SNAC_FAMILY_CTL              = 0x0001
AIM_SNAC_CTL_ERROR               = 0x0001
AIM_SNAC_CTL_C_READY             = 0x0002    # client online and ready
AIM_SNAC_CTL_S_READY             = 0x0003    # server is ready
AIM_SNAC_CTL_C_REQ_NEW_SERVICE   = 0x0004    # client requests for new service
AIM_SNAC_CTL_S_REDIRECT          = 0x0005    # server redirects (response on x4)
AIM_SNAC_CTL_C_REQ_RATE          = 0x0006    # client request rate information
AIM_SNAC_CTL_S_RATE_RESPONSE     = 0x0007    # server rate respone
AIM_SNAC_CTL_C_RATE_ACK          = 0x0008    # rate response ack
AIM_SNAC_CTL_S_RATE_CHANGED      = 0x000a    # rate information changed
AIM_SNAC_CTL_S_PAUSE             = 0x000b
AIM_SNAC_CTL_S_RESUME            = 0x000d
AIM_SNAC_CTL_C_REQ_SCREEN_INFO   = 0x000e    # client request sreen info
AIM_SNAC_CTL_S_SCREEN_INFO       = 0x000f
AIM_SNAC_CTL_S_EVIL              = 0x0010    # evil notification
AIM_SNAC_CTL_S_MIGRATION         = 0x0012    # migration notice/request
AIM_SNAC_CTL_S_MSG_OF_THE_DAY    = 0x0013    # message of the day
AIM_SNAC_CTL_C_SET_PRIVACY       = 0x0014    # client set the privacy flag
AIM_SNAC_CTL_S_URLS              = 0x0015    # server well known urls
AIM_SNAC_CTL_S_NOP               = 0x0016    # no operation
AIM_SNAC_CTL_S_HASH_REQUEST      = 0x001f    # server, hash request
AIM_SNAC_CTL_C_HASH_RESPONSE     = 0x0020    # client, hash response

AIM_SNAC_FAMILY_SERV_LOC         = 0x0002
AIM_SNAC_LOC_ERROR               = 0x0001
AIM_SNAC_LOC_C_REQ_PRIV          = 0x0002    # client request rights info
AIM_SNAC_LOC_S_PRIV              = 0x0003    # server rights response
AIM_SNAC_LOC_C_SET_USER_INFO     = 0x0004
AIM_SNAC_LOC_C_REQ_USER_INFO     = 0x0005
AIM_SNAC_LOC_S_USER_INFO         = 0x0006
AIM_SNAC_LOC_C_WATCHER_REQ       = 0x0007    # client watcher sub request
AIM_SNAC_LOC_S_WATCHER_NOTIFY    = 0x0008    # server watcher notification

AIM_SNAC_FAMILY_LIST             = 0x0003
AIM_SNAC_LST_ERROR               = 0x0001
AIM_SNAC_LST_C_REQ_PRIV          = 0x0002
AIM_SNAC_LST_S_PRIV              = 0x0003
AIM_SNAC_LST_C_ADD               = 0x0004    # client add buddy to buddy list
AIM_SNAC_LST_C_DEL               = 0x0005
AIM_SNAC_LST_C_LIST_QUERY        = 0x0006    # watcher list query
AIM_SNAC_LST_S_LIST              = 0x0007    # watcher list response
AIM_SNAC_LST_WATCHER_REQ         = 0x0008    # watcher sub request
AIM_SNAC_LST_S_WATCHER_NOTIFY    = 0x0009
AIM_SNAC_LST_S_REJECT_NOTIFY     = 0x000a    # server reject notification
AIM_SNAC_LST_S_IN                = 0x000b    # oncoming buddy
AIM_SNAC_LST_S_OUT               = 0x000c    # outgoing buddy

AIM_SNAC_FAMILY_MSG              = 0x0004
AIM_SNAC_MSG_ERROR               = 0x0001
AIM_SNAC_MSG_C_ADD_PAR           = 0x0002    # client add ICBM parameter
AIM_SNAC_MSG_C_REM_PAR           = 0x0003    # client remove ICBM par
AIM_SNAC_MSG_C_REQ_PAR           = 0x0004    # client request parameter
AIM_SNAC_MSG_S_INFO_PAR          = 0x0005    # server parameter information
AIM_SNAC_MSG_C_MSG               = 0x0006    # message from client
AIM_SNAC_MSG_S_MSG               = 0x0007    # message to the client
AIM_SNAC_MSG_C_EVIL_REQ          = 0x0008    # client evil request
AIM_SNAC_MSG_S_EVIL_REPLY        = 0x0009    # server evil reply
AIM_SNAC_MSG_S_MISSED            = 0x000a    # server missed call
AIM_SNAC_MSG_CS_ERROR            = 0x000b    # client error
AIM_SNAC_MSG_S_HOST_ACK          = 0x000c    # server host ack

AIM_SNAC_FAMILY_BOS              = 0x9
AIM_SNAC_BOS_ERROR               = 0x0001
AIM_SNAC_BOS_C_REQ_PRIV          = 0x0002    # client request BOS rights
AIM_SNAC_BOS_S_PRIV              = 0x0003    # server BOS rights response
AIM_SNAC_BOS_C_SET_GPERM         = 0x0004    # client set group permission mask
AIM_SNAC_BOS_C_ADD_PERM_ENTRY    = 0x0005    # client add perm. list entries
AIM_SNAC_BOS_C_DEL_PERM_ENTRY    = 0x0006    # client delete perm. list entries
AIM_SNAC_BOS_C_ADD_DENY_ENTRY    = 0x0007    # client add deny entry
AIM_SNAC_BOS_C_DEL_DENY_ENTRY    = 0x0008    # client delete deny list entries
AIM_SNAC_BOS_S_ERROR             = 0x0009    # server BOS error



# TLV codes
TLV_UIN        = 0x01
TLV_PWD        = 0x02
TLV_REDIR      = 0x05
TLV_COOKIE     = 0x06



#
# strange hardcoded data 
#
AIM_RAW_RATE_ACK    = '\x00\x01\x00\x02\x00\x03\x00\x04'
AIM_RAW_READY        = '''\
\x00\x01\x00\x03\x00\x04\x06\x86\
\x00\x02\x00\x01\x00\x04\x00\x01\
\x00\x03\x00\x01\x00\x04\x00\x01\
\x00\x04\x00\x01\x00\x04\x00\x01\
\x00\x09\x00\x01\x00\x04\x00\x01\
\x00\x0a\x00\x01\x00\x04\x00\x01\
\x00\x0b\x00\x01\x00\x04\x00\x01\
'''
AIM_RAW_23        = '''\
\x00\x01\x00\x03\x00\x02\x00\x01\
\x00\x03\x00\x01\x00\x04\x00\x01\
\x00\x06\x00\x01\x00\x08\x00\x01\
\x00\x09\x00\x01\x00\x0a\x00\x01\
\x00\x0b\x00\x02\x00\x0c\x00\x01\
\x00\x13\x00\x01\x00\x15\x00\x01\
'''
AIM_RAW_SET_INFO    = '''\
\x00\x00\x00\x00\x00\x03\x1f\x40\x03\xe7\x03\xe7\x00\x00\x00\
'''
AIM_RAW_HASH        = '''\
\x00\x10\xd4\x1d\x8c\xd9\x8f\x00\
\xb2\x04\xe9\x80\x09\x98\xec\xf8\
\x42\x7e\
'''


AIM_TOOL_JAVA          = 0x0001
AIM_TOOL_MAC           = 0x0002
AIM_TOOL_WIN16         = 0x0003
AIM_TOOL_WIN32         = 0x0004
AIM_TOOL_MAC68K        = 0x0005
AIM_TOOL_MACPPC        = 0x0006
AIM_TOOL_NEWWIN        = 0x0010







    


# Python doesn't have class methods,
# and it's class attributes is read-only :-(
# do this crap to implement counter used as next-sequence-number
# for the FLAP packets
__flap_sequence = 0
def inc_flap_sequence():
    new = __flap_sequence + 1
    globals()['__flap_sequence'] = new
    return new



# the same for SNAC class
__snac_sequence = 0
def inc_snac_sequence():
    new = __snac_sequence + 1
    globals()['__snac_sequence'] = new
    return new



class CProto:
    '''
        Generic protocol abstraction class
        proto_fmt  - format string to struct.[pack|unpack],
            to extract|fill data from|to raw  packet

        proto_fields - protocol fields, should correspond
            to format string

        proto_headerlen - length of the protocol fields
            excluding data length 

        proto_proto - identifier of underlyuing protocol
    '''
    proto_fmt    = None
    proto_fields    = None
    proto_headerlen    = None
    proto_proto    = None


    def  __init__(self, raw_data=None, *initializers):
        self.packet    = raw_data
        if raw_data is not None:
            fmt = self.proto_fmt % (len(raw_data) - 
                self.proto_headerlen)
            unpacked = struct.unpack(fmt, raw_data)
            map(lambda x,y,s=self:setattr(s,x,y),
                self.proto_fields, unpacked)
        else:
            # substitute appropriate sequence number
            # the position for sequence number
            # is different for FLAP and SNAC,
            # so substitute carefully
            if self.proto_proto == 'FLAP':
                sequence = inc_flap_sequence()
            elif self.proto_proto == 'SNAC':
                sequence = inc_snac_sequence()
            i = list(initializers)
            i.insert(self.proto_fields.index('sequence'), sequence)

            # set attributes
            map(lambda x,y,s=self:setattr(s,x,y),self.proto_fields, i)

            # generate raw packet
            args    = [self.proto_fmt % len(self.data)]
            args.extend(i)
            self.packet = apply(struct.pack, args)



class CFLAP(CProto):
    '''
        FLAP protocol abstraction
        this is the 'transport' for the SNAC frames
    '''

    proto_fmt    = '!BBHH %ds'
    proto_fields    = ['_unused', 'channel', 'sequence', 'data_len', 'data']
    proto_headerlen    = 0x6
    proto_proto    = 'FLAP'

    # overload constructor, to set correct data field
    def __init__(self,  raw_data=None, *initializers):
        apply(CProto.__init__, [self, raw_data] + list(initializers))
        self.data = self.data[:self.data_len]


class CSNAC(CProto):
    '''
        SNAC protocol abstraction
        this is the core of Oscar protocol.
        look at the definitions for family and
        subtype fields above
    '''

    proto_fmt    = '!HHBBL %ds'
    proto_fields    = ['family', 'subtype', 'flags0',
                'flags1', 'sequence', 'data']
    proto_headerlen    = 0xa
    proto_proto    = 'SNAC'



class CMsgOut:
    "class for outgoing message"

    #fmt = '!LL HB %ds HH B HHHHHH %ds'
    fmt = '!LLH B %ds H H 7s  H  L %ds'

    def __init__(self, uin, msg):
        u_l = len(uin)
        m_l = len(msg)

        fmt = self.fmt % (u_l, m_l)
        self.packet = struct.pack(fmt,
            0xfee00d00, 0x754b0000, 1,
            u_l,
            uin,
            2,
            m_l + 0xd,
            '\x05\x01\x00\x01\x01\x01\x01',
            m_l + 0x4,
            0x0,
            msg )

class CMsgIn:
    "class for incoming message"

    #fmt = '!10s B %ds HHHHHHH L HH L HHHHH BBBBBBB H L %ds'
    fmt = '!10s B %ds %ds %ds'

    def __init__(self, data):
        '''
            construct CMsgIn object
            from incoming dat
        '''

        # data length
        d_l = len(data)

        # calculate uin length and message length
        u_l    = struct.unpack('B', data[0x0a])[0]

        # message len we can get
        # searching data for null terminator
        # starting from the end
        m_l = d_l - data.rfind('\0')

        # make correct format string
        fmt  = self.fmt % (u_l, d_l - u_l - m_l - 0xb , m_l)

        # something wrong, log it
        if m_l <= 0 or d_l <= 0:
            log.logger.log_dump(data)
            self.uin = self.msg = None
            return

        # unpack the data,
        # u is the tuple that corresponds to format string
        u = struct.unpack(fmt, data)
        self.uin    = u[2]
        self.msg    = u[4]



class CBuddyOnline:
    '''
        class for 'oncoming buddy' message
    '''

    fmt = '!10s B %ds %ds'

    def __init__(self, data):
        # entire packet length
        d_l = len(data)
        # uin len
        u_l = struct.unpack('B', data[0x0a])[0]
        # rest of message length
        m_l = d_l - 0xb - u_l
        fmt = self.fmt % (u_l, m_l )

        # unpack the packet
        u = struct.unpack(fmt, data)
        self.uin_len, self.uin = u[1], u[2]






