# $Id: ui_tk.py,v 1.7 2001/06/18 09:43:53 drozd Exp $
# (c) 2001 Sergey M Lyubka <devnull@tecnomen.ie>
#
# User Interface file




#interface with main module as following:
#
#class CUI:
#
#    def __init__(self, session, label):
#        """     session - instance of CICQSession
#        """
#
#
#    def get_input(self):
#        """
#            called by CICQSession instance.
#            should return tuple (cmd,args)
#            where cmd is a valid cmd from cmd_table
#        """
#
#    def contact_add(self, buddy):
#        """
#            adds a new buddy to the contact list
#            buddy is a CBuddy instance
#            (take a look on buddy.py)
#        """
#
#    def set_contact_mode(self, uin, mode=None):
#        """
#            set mode for UIN
#        """
#
#    def msg_in(self, from, to, msg)
#        """
#            message: from --> to
#            from, to are the UINs
#        """
#
#    def get_passwd(self)
#        """
#              should return password string (clear text)
#        """


from Tkinter import *


import struct
import string
import select
import const
import sys
import threading
import time

import config
import buddy


COLOR_ONLINE    = 'blue'
COLOR_OFFLINE    = 'red'


ROOT_DIR    = config.get_root_dir()


# Tk initialization
r        = Tk()

IMG_PROP_BUTTON = PhotoImage(file = ROOT_DIR + '/images/leaf.gif')
IMG_ONLINE      = PhotoImage(file = ROOT_DIR + '/images/online.gif')
IMG_OFFLINE     = PhotoImage(file = ROOT_DIR + '/images/offline.gif')


def cleanup():
    "cleanup callback"
    print 'quitting GUI...'
    r.session.quit = 1
    r.quit()


class CUI:
    '''
        GUI abstraction
    '''

    def __init__(self, session):
        '''User Interface constluctor
            session is a reference to CICQSession instance
        '''

        self.session    = session
        self.uin        = session.uin
        self.contacts   = session.contacts

        # empty the command queue
        self.cmd_queue    = []


        # dictionary contains messages from every contact
        # key in the contact's uin
        self.messages    = {}

        # init message list
        self.messages[self.uin] = []

        self.gui_init()



    def gui_init(self):
        # Tk shit
        r.title(self.uin)
        r.resizable(NO,YES)
        r.protocol("WM_DELETE_WINDOW", cleanup)
        # glue for quit
        r.session    = self.session

        main = self.main    = Frame()
        frame_top        = Frame(main)
        frame_bottom        = Frame(main)

        # contact list
        # second argument = we are
        self.contact_list    = CContactList(frame_top, self,
            fg=COLOR_OFFLINE)

        # properties button
        prop    = self.prop =CPropButton(self, frame_bottom, 
            image=IMG_PROP_BUTTON)
        self.status_bar    = CStatusBar(r)

        main.pack(fill=BOTH, expand=YES)
        frame_top.pack(fill=BOTH, expand=YES)
        frame_bottom.pack(side=BOTTOM, fill=BOTH)
        self.contact_list.pack(fill=BOTH, expand=YES)
        prop.pack(side=LEFT)
        self.status_bar.pack(fill=X, side=BOTTOM)

        threading.Thread(target=r.mainloop).start()


    def get_input(self):
        ''' 
            we are called by icq session
            and should return list of commands:
            [(cmd1, params1), (cmd2, params2)]
            cmd<n> should present in cmd_table 
            (look const.py)
        '''
        ret_val = self.cmd_queue[:]
        self.cmd_queue = []
        return ret_val


    def contact_add(self, b):
        "add a contact to the contact list"
        self.contacts[repr(b)] = b
        self.contact_list.add(b)
        self.messages[b.uin]    = []    # create an emty message list
                        # for this contact


    def set_contact_mode(self, uin, mode):
        "changes the mode for the constact"

        nick = self.lookup_contact_by_uin(uin).nick
        # that guy isn't known
        if nick is None:
            nick = uin
            self.contact_add(buddy.CBuddy(uin,nick))
        # change the icon
        self.contact_list.set_mode(nick, mode)



    def add_cmd(self, cmd, *params):
        "queue command for execution"
        self.cmd_queue.append((cmd,params))


    def msg_in(self, src, dst, msg):
        '''
            handles received messages
            update a dialog window if necessary
            create it if it doesn't exists
        '''

        lb = self.contact_list
        self.messages[src].append(CMessage(src, dst, msg))

        # look which dialog to update
        target = (dst == self.uin and [src] or [dst])[0]
        contact = self.lookup_contact_by_uin(target)

        # this guy is unknown for us!
        # add him with some 'reasonable' nickname :-)
        # nickname will be string representation for the uin
        # XXX: should be replaced with something sane !
        if contact is None:
            b = buddy.CBuddy(src, src)
            self.contact_add(uin, b)
            contact = b

        if target not in lb.dialogs.keys():
            wnd    = CMessageWindow(lb, lb.ui, contact.nick)
            lb.dialogs[target] = wnd
        else:
            self.contact_list.dialogs[target].update_dialog()



    def lookup_contact_by_nick(self, nick):
        "return buddy oject for given nickname"

        return self.contacts[nick]
        #retval = None
        #if nick in self.contacts.keys():
        #    retval =  self.contacts[nick]
        #return retval


    def lookup_contact_by_uin(self, uin):
        "return buddy oject for given uin"

        retval = None
        for contact in self.contacts.values():
            if contact.uin == uin:
                retval = contact
                break
        return retval


    def get_passwd(self):
        '''
            prompt for password and return the input
        '''

        def ok(ev):
            w = ev.widget.master
            # make bogus attribute to return the value
            w.master.pwd = w.t_entry.get(0.0, END)
            w.destroy()

        # fire up small dialog
        # pass top level frame as parent
        # to allow return value
        d = Toplevel(self.main)
        d.t_entry = Text(d)
        d.t_entry.pack(fill=BOTH)
        b_ok = Button(d, text='ok')
        b_ok.bind('<Button-1>', ok)
        b_ok.pack()

        # wait until password will be entered
        # in child dialog
        while 1:
            if not self.main.__dict__.has_key('pwd'):
                time.sleep(const.SELECT_TIMEOUT)
                continue
            return self.main.__dict__['pwd']
        #return raw_input('password: ')



class CMessage:
    " message abstraction"

    def __init__(self, src, dst, data):
        '''
            src - source uin
            dst - dsetination uin
            data - message's text
        '''

        self.src    = src    # from whom 
        self.dst    = dst    # to whom
        self.data    = data
        self.time    = time.time()



class CContactList(Text):
    "listbox that shows the contact list"

    # width of contact list, in letters
    width = 20

    def __init__(self, master, ui, cnf={}, **kw):
        '''
            ui is CUI class instance
        '''

        frame = Frame(master)#, cnf)
        self.master = master
        self.ui = ui    # this is top level instance
        Text.__init__(self, frame, relief=SUNKEN, 
            state=DISABLED, wrap=WORD, width=self.width)
        scroll = Scrollbar(frame, relief=SUNKEN, command=self.yview)
        scroll.pack(side=RIGHT, fill=Y)
        self.pack(fill=BOTH, expand=YES, side=RIGHT)
        self['yscrollcommand'] = scroll.set

        # bind mouse buttons
        self.tag_config('tag_click')
        self.tag_bind('tag_click', "<Double-Button-1>", create_dialog)

        # redefine "pack", "place" and "grid" 
        #to be the frame instance's methods

        self.pack =  frame.pack
        self.place = frame.place
        self.grid = frame.grid

        # this is a list of opened send message windows
        # keyed by UINs
        self.dialogs = {}


    def add(self, b):
        '''
            adds a contact to the contact list
            b is CBuddy instance
        '''

        self.config(state=NORMAL)
        self.image_create(END, image=IMG_OFFLINE)
        self.insert(END, ' %s\n' % b.nick, 'tag_click')
        self.config(state=DISABLED)


    def get_nick_by_index(self, lineno):
        '''
            return nick for given line number or None
        '''

        nick = self.get('%s.2' % lineno, '%s.0' % (lineno + 1))
        if len(nick) == 0:
            return None

        # strip '\n' from the end
        return nick[:-1]



    def get_index_by_nick(self, nick):
        '''
            return line number for given nick or None
        '''

        for i in range(1, len(self.ui.contacts)):
            n = self.get_nick_by_index(i)
            if n == nick:
                return i

        return None


    def set_mode(self, nick, mode):
        '''
            change contact's mode
        '''
        # lookup listbox index for given nick
        # XXX: can be optimized

        # change the icon
        img = (mode==const.ICQ_MODE_ONLINE and [IMG_ONLINE] or [IMG_OFFLINE])[0]

        i = self.get_index_by_nick(nick)
        if i is None:
            print 'something wrong in set_mode'
            return
        index = '%d.0' % i
        self.config(state=NORMAL)
        self.delete(index)
        self.image_create(index, image=img)
        self.config(state=DISABLED)




def create_dialog(event):
    '''
        create new dialog window function
        called by event on listbox widget,
        normally by doble-click on contact nickname
        add an entry of open dialogs
    '''
    list  = event.widget
    # extract the nickname
    # XXX: replace with something sane
    nick = None
    for i in range(1,len(list.ui.contacts)):
        info = list.dlineinfo('%s.0' % i)
        if info is not None and \
            info[1] + info[3] > event.y and info[1] < event.y:

            # this is the index
            nick = list.get_nick_by_index(i)

    # nothing found
    if nick is None:
        return

    # lookup the uin
    uin   = list.ui.lookup_contact_by_nick(nick).uin

    # create new dialog window if necessary
    if list.dialogs.has_key(uin):
        return
    else:
        wnd   = CMessageWindow(list, list.ui, nick)
        list.dialogs[uin] = wnd



class CMessageWindow(Toplevel):
    "Message window abstraction"

    up_height    = 7    # height of t_dialog window, lines
    down_height  = 2    # height of t_input window, lines
    width        = 40   # width of message window, letters

    def __init__(self, master, ui, nick, **kw):
        '''
            dialog window constructor
            master - listbox widget, with contact list
            ui - our 'main' GUI engine
            nick is a target nickname
        '''

        Toplevel.__init__(self, borderwidth=10)
        self.title("Message: %s" % nick)
        self.master     = master
        self.ui     = ui
        self.contact    = self.ui.lookup_contact_by_nick(nick)

        # register destroy handler
        self.protocol('WM_DELETE_WINDOW', self.destroy)

        f_main    = Frame(self)
        f_up    = Frame(f_main)
        f_down    = Frame(f_main, height=20)
        # make bottom frame with button notresizable
        f_down.resizable = (NO, NO)

        self.t_dialog= Text(f_up, borderwidth=5,
            wrap=WORD,state=DISABLED,
            width=self.width, height=self.up_height)# show dialog 
        self.t_dialog.tag_config('tag_red', foreground='red')
        self.t_dialog.tag_config('tag_blue', foreground='blue')

        self.t_input = Text(f_up, borderwidth=5,
            wrap=WORD, height=self.down_height,
            width=self.width)    # here we type

        scroll = Scrollbar(f_up, 
            relief=SUNKEN, command=self.t_dialog.yview)
        self.t_dialog['yscrollcommand'] = scroll.set

        self.t_input.bind('<Control-Return>',
            lambda ev: ev.widget.master.master.send)
        self.t_input.send = self.send

        # create 'send' and 'cancel' buttons
        b_send   = Button(f_down, text='send')
        b_exit   = Button(f_down, text='exit')
        b_send.bind('<Button-1>', lambda x:x.widget.master.master.master.send())
        b_exit.bind('<Button-1>', lambda x:x.widget.master.master.master.destroy())
        b_send.pack(side=RIGHT)
        b_exit.pack(side=RIGHT)


        f_main.pack(fill=BOTH, expand=YES)
        f_up.pack(fill=BOTH, expand=YES)
        f_down.pack(fill=BOTH, expand=NO)
        scroll.pack(side=RIGHT, fill=Y)
        self.t_dialog.pack(fill=BOTH, expand=YES)
        self.t_input.pack(fill=BOTH, expand=YES)

        self.update_dialog()


    def destroy(self):
        "destroy handler"

        # delete window from the dict
        del self.master.dialogs[self.contact.uin]
        Toplevel.destroy(self)


    def update_dialog(self):
        "updates dialog window "

        # build list of messages , in and out,
        # sort it and render in dialog window
        target_uin = self.contact.uin
        my_msgs = self.ui.messages[self.ui.uin]
        messages = filter(lambda m, t=target_uin: m.dst == t, my_msgs)
        messages.extend(self.ui.messages[target_uin])
        messages.sort(lambda x,y:(x.time > y.time and [1] or [-1])[0])

        d = self.t_dialog

        d.config(state=NORMAL)
        d.delete(0.0, END)
        for msg in messages:
            if msg.src == self.ui.uin:
                tag = 'tag_blue'
            else:
                tag = 'tag_red'

            t = time.strftime('%H.%M', time.localtime(msg.time))
            sender = self.ui.lookup_contact_by_uin(msg.src)
            d.insert(0.0, '%s\n' % msg.data)
            d.insert(0.0, '(%s) %s: ' % (t,sender), tag)
        self.t_dialog.config(state=DISABLED)
        del messages


    def send(self):
        '''
            queue message for sending by icq session
            update appropriate dialog window
        '''
        # get text entered
        txt = self.t_input.get(0.0,END)
        # if nothing entered
        if len(string.split(txt)) == 0:
            return
        txt = string.rstrip(string.lstrip(txt))

        # add message to the list
        self.ui.msg_in(self.ui.uin, self.contact.uin, txt)

        # queue for sending
        self.ui.add_cmd('msg-out', self.contact.uin, txt)

        # update dialog window :
        self.update_dialog()

        # clear input window
        self.t_input.delete(1.0, END)




class CPropButton(Menubutton):
    "a button with cool FreeBSD logo"

    def __init__(self, ui, master=None, **kw):
        self.ui    = ui
        Menubutton.__init__(self, master, kw, 
            direction='right', relief=RAISED)

        menu = self['menu'] = Menu(self)

        menu.add_command(label='Find user')
        menu.add_command(label='Options')
        menu.add_command(label='Quit', command=cleanup)




class CStatusBar(Frame):

    def __init__(self, master):
        Frame.__init__(self, master)
        self.label = Label(self, text='Status: ',
            bd=1, relief=SUNKEN, anchor=W)
        self.label.pack(fill=X)

    def set(self, format, *args):
        self.label.config(text=format % args)
        self.label.update_idletasks()

    def clear(self):
        self.label.config(text="")
        self.label.update_idletasks()
